/*
 *
 * To compile the test app (main()):
 * gcc -Wall access_config_configfile.c xalloc.c -DTEST_APP=1
 *
 * */

#include "xalloc.h"
#include "access_config.h"

#include <assert.h>
#include <stdio.h>
#include <inttypes.h>
#include <string.h>
#include <netinet/in.h>

// VIOD_ID IOD_ID IP_ADDR IP_PORT CONFIG_PATH
#define MAX_IP_ADDR_LEN 20
#define MAX_CONFIG_PATH_LEN 255

#define MAX_TEMP_IOD 512
static struct temp_iod_t * temp_iods[MAX_TEMP_IOD];

// The structure containing a config file line
struct temp_iod_t {
	int viod_id;
	int iod_id;
	uint32_t ipv4;
	int ip_port;
	char *config_path;	// malloc'ed
};

// Allocate a temp_iod_t structure
static struct temp_iod_t *
new_temp_iod_t()
{
	struct temp_iod_t *tmp;
	tmp = xmalloc(sizeof (*tmp));
	return tmp;
}

// Free a new_temp_iod_t() allocated structure
static void
destroy_temp_iod_t(struct temp_iod_t *t)
{
	if (t) {
		xfree(t->config_path);
	}
	xfree(t);
}

// Returns a new_temp_iod_t()'ed temp_iod_t based on what's in line
static struct temp_iod_t *
parseline(const char * const line)
{
	int ret;
	struct temp_iod_t *t;
	char *ip_addr;

	//printf("DBG: parsing [%s]\n", line);

	t = new_temp_iod_t();

	// Note: %as matches a string and malloc()'s it
	ret = sscanf(line, "%d %d %as %d %as", &t->viod_id, &t->iod_id, &ip_addr, &t->ip_port, &t->config_path);
	if (ret != 5) {
		printf("Error: wrong line format [%s]\n", line);
		xfree(t);
		return NULL;
	}

	{
		int i;
		int ip[4];
		uint32_t ipv4;
		ret = sscanf(ip_addr, "%d.%d.%d.%d", ip, ip + 1, ip + 2, ip + 3);
		if (ret != 4) {
			printf("Error: the IP address must match n.n.n.n with each n in 0..255\n");
			xfree(t);
			return NULL;
		}
		ipv4 = 0;
		for (i = 0; i < sizeof (ip) / sizeof (*ip); i++) {
			printf("ip[%d] == %d == %u\n", i, ip[i], (unsigned int) ip[i]);
			if (ip[i] < 0 || ip[i] > 255) {
				printf("Error: the IP address must match n.n.n.n with each n in 0..255\n");
				xfree(t);
				return NULL;
			}
			ipv4 = (ipv4 << 8) | ((unsigned int) ip[i]);
		}

		t->ipv4 = ntohl(ipv4);

	}

	//printf("DBG: parsing done\n");
	return t;
}

// This function reads filename and fills temp_iods[]
// Returns:
//   SUCCESS: the number of lines parsed
//   FAILURE: -1
static int get_temp_iods(const char * const filename) {
	
	int idx = 0;
	char line[80];
	FILE *f;

	f = fopen(filename, "r");
	if (!f) {
		printf("Error: cannot open [%s]\n", filename);
		return -1;
	}

	while (!feof(f)) {
		if (!fgets(line, sizeof(line), f)) {
			if (feof(f)) {
				goto out;
			}
			printf("Error: line %d\n", idx+1);
			goto error_out;
		}
		temp_iods[idx] = parseline(line);
		if (!temp_iods[idx]) {
			printf("Error: line %d\n", idx+1);
			goto error_out;
		}
		idx++;
		if (idx >= MAX_TEMP_IOD) {
			printf("Error: too many iods (limit is %d)\n", MAX_TEMP_IOD);
			goto error_out;
		}
	}

out:
	return idx;

error_out:
	fclose(f);
	return -1;
}

int get_config(struct viod_t **viodsp, int *viods_sizep) {

	int i;
	int nb_temp_iods;

	struct viod_t *viods;
	int viods_size;

	nb_temp_iods = get_temp_iods("cfg");
	if (nb_temp_iods > 0) {
		printf("INFO: found %d temp_iods\n", nb_temp_iods);
	}

	viods = NULL;
	viods_size = 0;
	for(i = 0; i < nb_temp_iods; i++) {
		int old_nbiods;
		int viod_id = temp_iods[i]->viod_id;
		printf("iod_id = %d viod_id = %d\n", temp_iods[i]->iod_id, viod_id);

		if (viod_id + 1 > viods_size) {
			int j;
			int old_viods_size = viods_size;
			
			viods_size = 1 + viod_id;
			printf("Increasing viods[] size from %d to %d\n", old_viods_size, viods_size);
			viods = realloc(viods, viods_size * sizeof(*viods));
			if (!viods) {
				printf("Error: realloc(viods,...) failed\n");
				exit(1);
			}
			// Zero the newly allocated data
			for (j = old_viods_size; j < viods_size ; j++) {
				printf("Zeroing viods[%d]\n", j);
				viods[j].id = j;
				viods[j].nbiods = 0;
				viods[j].iods = NULL;
			}
			printf("Zeroing done\n");
		}


		//assert(viods[viod_id] == viod_id);
	
		// The viods[] size is now increased, and the fields in the
		// last entry are set to the current temp_iods[i] record
		old_nbiods = viods[viod_id].nbiods;

		viods[viod_id].iods = realloc(viods[viod_id].iods, (old_nbiods + 1) * sizeof(viods[viod_id].iods[0]));
		if (!viods[viod_id].iods) {
			printf("Error: realloc(viods[viod_id].viods,...) failed\n");
			exit(1);
		}
		viods[viod_id].iods[old_nbiods].id = temp_iods[i]->iod_id;
		memset(&viods[viod_id].iods[old_nbiods].sin, '\0', sizeof(viods[viod_id].iods[old_nbiods].sin));
		viods[viod_id].iods[old_nbiods].sin.sin_family = AF_INET;
		viods[viod_id].iods[old_nbiods].sin.sin_port = temp_iods[i]->ip_port;
		viods[viod_id].iods[old_nbiods].sin.sin_addr.s_addr = temp_iods[i]->ipv4;

		viods[viod_id].nbiods++;

		destroy_temp_iod_t(temp_iods[i]);
	}


	// TODO: add some checks to ensure:
	//   - no twice the same IOD id
	//   - no twice the same sin

	*viodsp = viods;
	*viods_sizep = viods_size;
		
	return 0;
}

void
free_config(struct viod_t *viods, int viods_size)
{
	int i;
	if (viods) {
		for (i = 0; i < viods_size; i++) {
			xfree(viods[i].iods);
		}
	}
	xfree(viods);
}


#ifdef TEST_APP
#warning "TEST_APP enabled"

int
main(int argc, char *argv[])
{
	char *lines[] = {
		"1 11 1.2.3.4 1214 /foo/bar1",
		"2 12 1.2.3.5 1214 /foo/bar2",
		"2 13 1.2.3.6 1214 /foo/bar3",
	};
	
	//{	"1 2 42 1.2.3.4 1214 /foo/bar2",
	//	"12 42 1.2.3.4 12a14 /foo/bar3",
	//	"foo bar4",
	//};
	struct temp_iod_t *t;
	int i;

	printf("\n--------- test 1 ---------\n");
	printf("--------- test 1 ---------\n");
	for (i = 0; i < sizeof (*lines) / (sizeof **lines); i++) {
		char *line = lines[i];

		printf("\n--------- line %d ---------\n", i);

		t = parseline(line);
		if (!t) {
			printf("Error: parseline failed\n");
			continue;
		}

		printf("viod_id=%d iod_id=%d ip_addr=0x%08x ip_port=%d config_path=[%s]\n", t->viod_id, t->iod_id, t->ipv4, t->ip_port, t->config_path);

		destroy_temp_iod_t(t);
	}

	printf("\n--------- test 2 ---------\n");
	printf("--------- test 2 ---------\n");
	{
		int i, j;
		struct viod_t *viods;
		int viods_size;
		get_config(&viods, &viods_size);

		for (i = 0; i < viods_size; i++) {
			printf("VIOD = %d\n", i);
			for (j = 0; j < viods[i].nbiods; j++) {
				printf("  IODS[%d] = %d\n", j, viods[i].iods[j].id);
			}
		}

		free_config(viods,  viods_size);
		
	}
	return 0;
}
#endif
